/*
 *  EXPRESSION, {akii-root}/src/expression.c
 *
 * ==========================================================================
 *
 *                      Arena Kliens II. Releases
 *
 * ==========================================================================
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * A legkiralyabb szovegszerkeszto: VIM - Vi IMproved 5  <http://www.vim.org>
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "expression.h"
#include "variables.h"
#include "commands.h"
#include "error.h"
#include "debug.h"

#define OPNUM	        27
#define OPDUPNUM        15

#define DF(f)   (f-(long)f)

char *op_eq(struct kif *kif);
char *op_plus_eq(struct kif *kif);
char *op_minus_eq(struct kif *kif);
char *op_times_eq(struct kif *kif);
char *op_slash_eq(struct kif *kif);
char *op_mod_eq(struct kif *kif);
char *op_cat_eq(struct kif *kif);
char *op_or(struct kif *kif);
char *op_and(struct kif *kif);
char *op_is_eq(struct kif *kif);
char *op_is_noteq(struct kif *kif);
char *op_less(struct kif *kif);
char *op_greater(struct kif *kif);
char *op_less_eq(struct kif *kif);
char *op_greater_eq(struct kif *kif);
char *op_plus(struct kif *kif);
char *op_minus(struct kif *kif);
char *op_cat(struct kif *kif);
char *op_times(struct kif *kif);
char *op_slash(struct kif *kif);
char *op_slash_egesz(struct kif *kif);
char *op_mod(struct kif *kif);
char *op_u_plus(struct kif *kif);
char *op_u_minus(struct kif *kif);
char *op_u_not(struct kif *kif);
char *op_u_plusplus(struct kif *kif);
char *op_u_minusminus(struct kif *kif);

static struct operators oplist[OPNUM] =
{
    "+=", 0, op_plus_eq,
    "-=", 0, op_minus_eq,
    "*=", 0, op_times_eq,
    "/=", 0, op_slash_eq,
    "%=", 0, op_mod_eq,
    "~=", 0, op_cat_eq,
    "||", 1, op_or,
    "&&", 2, op_and,
    "==", 3, op_is_eq,
    "!=", 3, op_is_noteq,
    "<=", 4, op_less_eq,
    ">=", 4, op_greater_eq,
    "++", PL_BIUNARY, op_u_plusplus,
    "--", PL_BIUNARY, op_u_minusminus,
    "/%", 6, op_slash_egesz,
    "=", 0, op_eq,
    "<", 4, op_less,
    ">", 4, op_greater,
    "+", 5, op_plus,
    "-", 5, op_minus,
    "~", 5, op_cat,
    "*", 6, op_times,
    "/", 6, op_slash,
    "%", 6, op_mod,
    "+", 7, op_u_plus,
    "-", 7, op_u_minus,
    "!", 7, op_u_not,
};

int isopchar(char c);
char *getoperand(struct kif *kif, int *mustfree);

char *dokif(struct kif *kif)
{
    char *s, *p;
    int mustfree;

#ifdef AKII_DEBUG
    dbgenter(dbg_dokif, "?");
#endif
    if (!kif) {
#ifdef AKII_DEBUG
	dbgleave(dbg_dokif);
#endif
	return (NULL);
    }
    s = getoperand(kif, &mustfree);
    if (!s) {
#ifdef AKII_DEBUG
	dbgleave(dbg_dokif);
#endif
	return (NULL);
    }
    if (!mustfree) {
	p = malloc(strlen(s) + 1);
	strcpy(p, s);
#ifdef AKII_DEBUG
	dbgleave(dbg_dokif);
#endif
	return (p);
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_dokif);
#endif
    return (s);
}
struct kif *readkif(char **s)
{
    struct kif *kif;

#ifdef AKII_DEBUG
    dbgenter(dbg_readkif, *s);
#endif
    kif = readkifinfix(s);
    kif = mkkiffa(kif);
#ifdef AKII_DEBUG
    dbgleave(dbg_readkif);
#endif
    return (kif);
}
struct kif *newkif(enum kifenum tipus)
{
    struct kif *kif;
    kif = malloc(sizeof(struct kif));

    kif->prev = kif->next = NULL;
    kif->szulo = kif->bal = kif->jobb = NULL;
    kif->tipus = tipus;
    return (kif);
}
struct kif *readkiffv(char **s, enum peeenum *pee, char *p, char *temp)
{
    struct kif *kif;
    char TEMP[222];
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_readkiffv, *s);
#endif
    if (*pee >= PEE_VPRE) {
	errno = EWRONGKIF;
	errtemp[0] = 0;
#ifdef AKII_DEBUG
	dbgleave(dbg_readkiffv);
#endif
	return (NULL);
    }
    i = (*s) - p;
    strncpy(TEMP, p, i);
    TEMP[i] = 0;
    while (isspace(**s) && (**s))
	(*s)++;
    if (**s != '(') {
	if (*pee != PEE_NOTHING) {
	    errno = EEZFV;
	    errtemp[0] = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_readkiffv);
#endif
	    return (NULL);
	}
	strcpy(temp, *s);
	*s += strlen(temp);
    }
    else {
	(*s)++;
	braces(s, temp, '(', ')');
	if (**s != ')') {
	    errno = ENOZAROJEL;
	    strcpy(errtemp, temp);
#ifdef AKII_DEBUG
	    dbgleave(dbg_readkiffv);
#endif
	    return (NULL);
	}
	(*s)++;
    }
    kif = newkif(KIF_FV);
    kif->ertek = malloc(i + 1);
    strncpy(kif->ertek, p, i);
    kif->ertek[i] = 0;
    kif->param = malloc(strlen(temp) + 1);
    strcpy(kif->param, temp);
    *pee = PEE_CONST;
#ifdef AKII_DEBUG
    dbgleave(dbg_readkiffv);
#endif
    return (kif);
}
struct kif *readkifop(char **s, enum peeenum *pee, int *maxpl)
{
    struct operators *op = NULL;
    struct kif *kif;
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_readkifop, *s);
#endif
    for (i = 0; i < OPNUM; i++) {
	if (strncmp(*s, oplist[i].nev, (i < OPDUPNUM) ? 2 : 1) == 0) {
	    if (oplist[i].pl == 0) {
		if ((*pee != PEE_VAR) || (*maxpl > 0)) {
		    errno = EVARKELL;
		    errtemp[0] = 0;
#ifdef AKII_DEBUG
		    dbgleave(dbg_readkifop);
#endif
		    return (NULL);
		}
		op = oplist + i;
		*pee = PEE_OP;
		break;
	    }
	    if ((oplist[i].pl > 0) && (oplist[i].pl < PL_UNARY)) {
		if (*pee >= PEE_VAR) {
		    op = oplist + i;
		    *pee = PEE_OP;
		    break;
		}
		continue;
	    }
	    if (oplist[i].pl == PL_UNARY) {
		if (*pee > PEE_OP) {
		    errno = EWRONGKIF;
		    errtemp[0] = 0;
#ifdef AKII_DEBUG
		    dbgleave(dbg_readkifop);
#endif
		    return (NULL);
		}
		op = oplist + i;
		*pee = PEE_PRE;
		break;
	    }
	    if (oplist[i].pl == PL_BIUNARY) {
		if (*pee <= PEE_OP) {
		    op = oplist + i;
		    *pee = PEE_VPRE;
		    i = BUNARY_PRE;
		    break;
		}
		if (*pee == PEE_VAR) {
		    op = oplist + i;
		    *pee = PEE_CONST;
		    i = BUNARY_POST;
		    break;
		}
	    }
	}
	if (op)
	    break;
    }
    if (!op) {
	errno = EWRONGKIF;
	errtemp[0] = 0;
#ifdef AKII_DEBUG
	dbgleave(dbg_readkifop);
#endif
	return (NULL);
    }
    kif = newkif(KIF_OP);
    kif->ertek = (char *) op;
    kif->param = (char *) i;
    (*s) += (i < OPDUPNUM) ? 2 : 1;
    *maxpl = op->pl;
#ifdef AKII_DEBUG
    dbgleave(dbg_readkifop);
#endif
    return (kif);
}
struct kif *milyenkifez(char **s, enum peeenum *pee, int *maxpl)
{
    char temp[1024], *t = temp, *p, *q;
    struct valtozo *var;
    struct kif *kif, *kif2;
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_milyenkifez, *s);
#endif
    p = *s;
    if (isalpha(**s)) {
	var = getvarifp(&varlist, s);
	if (errno) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	q = *s;
	while ((isspace(*q)) && (*q))
	    q++;
	if ((!isopchar(*q)) && (*pee == PEE_NOTHING))
	    if (q != *s) {
		kif = readkiffv(s, pee, p, temp);
#ifdef AKII_DEBUG
		dbgleave(dbg_milyenkifez);
#endif
		return (kif);
	    }
	if (var) {
	    if ((var->tipus == VAR_FV) || (var->tipus == VAR_ALIAS))
		if ((*q == '(') || (q != *s) || (!*q)) {
		    kif = readkiffv(s, pee, p, temp);
#ifdef AKII_DEBUG
		    dbgleave(dbg_milyenkifez);
#endif
		    return (kif);
		}
	}
	else {
	    if (*q == '(') {
		kif = readkiffv(s, pee, p, temp);
#ifdef AKII_DEBUG
		dbgleave(dbg_milyenkifez);
#endif
		return (kif);
	    }
	}
	if (*pee >= PEE_VAR) {
	    errno = EWRONGKIF;
	    errtemp[0] = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	i = (*s) - p;
	*s = q;
	kif = newkif(KIF_VAR);
	kif->ertek = malloc(i + 1);
	strncpy(kif->ertek, p, i);
	kif->ertek[i] = 0;
	*pee = PEE_VAR;
#ifdef AKII_DEBUG
	dbgleave(dbg_milyenkifez);
#endif
	return (kif);
    }
    if (**s == '$') {
	if (*pee >= PEE_VPRE) {
	    errno = EWRONGKIF;
	    errtemp[0] = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	dollar(s, NULL);
	if (errno) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	kif = newkif(KIF_CONST);
	i = (*s) - p;
	kif->ertek = malloc(i + 1);
	strncpy(kif->ertek, p, i);
	kif->ertek[i] = 0;
	*pee = PEE_CONST;
#ifdef AKII_DEBUG
	dbgleave(dbg_milyenkifez);
#endif
	return (kif);
    }
    if (**s == '"') {
	if (*pee >= PEE_VPRE) {
	    errno = EWRONGKIF;
	    errtemp[0] = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	(*s)++;
	delquote = 1;
	braces(s, temp, '"', '"');
	delquote = 0;
	if (**s != '"') {
	    errno = ENOIDEZOJEL;
	    errtemp[0] = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	kif = newkif(KIF_CONST);
	kif->ertek = malloc(strlen(temp) + 1);
	strcpy(kif->ertek, temp);
	(*s)++;
	*pee = PEE_CONST;
#ifdef AKII_DEBUG
	dbgleave(dbg_milyenkifez);
#endif
	return (kif);
    }
    if (**s == '(') {
	if (*pee >= PEE_VPRE) {
	    errno = EWRONGKIF;
	    errtemp[0] = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	(*s)++;
	braces(s, temp, '(', ')');
	if (**s != ')') {
	    errno = ENOZAROJEL;
	    errtemp[0] = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	kif2 = readkif(&t);
	if (!kif2) {
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	kif = newkif(KIF_KIF);
	kif->ertek = (char *) kif2;
	(*s)++;
	*pee = PEE_CONST;
#ifdef AKII_DEBUG
	dbgleave(dbg_milyenkifez);
#endif
	return (kif);
    }
    if (isdigit(**s)) {
	if (*pee >= PEE_VPRE) {
	    errno = EWRONGKIF;
	    errtemp[0] = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	strtod(*s, &t);
	i = t - (*s);
	kif = newkif(KIF_CONST);
	kif->ertek = malloc(i + 1);
	strncpy(kif->ertek, *s, i);
	kif->ertek[i] = 0;
	*pee = PEE_CONST;
	(*s) += i;
#ifdef AKII_DEBUG
	dbgleave(dbg_milyenkifez);
#endif
	return (kif);
    }
    if (isopchar(**s)) {
	kif = readkifop(s, pee, maxpl);
#ifdef AKII_DEBUG
	dbgleave(dbg_milyenkifez);
#endif
	return (kif);
    }
    errno = EWRONGKIF;
    errtemp[0] = 0;
#ifdef AKII_DEBUG
    dbgleave(dbg_milyenkifez);
#endif
    return (NULL);
}
struct kif *readkifinfix(char **s)
{
    struct kif *kif = NULL, *akif = NULL, *ekif = NULL;
    enum peeenum pee = PEE_NOTHING;
    int maxpl = 0;
    char *eleje;

    eleje = *s;
#ifdef AKII_DEBUG
    dbgenter(dbg_readkifinfix, *s);
#endif
    while (**s) {
	while (isspace(**s) && (**s))
	    (*s)++;
	if (!**s)
	    break;
	akif = milyenkifez(s, &pee, &maxpl);
	if (!akif) {
	    if (ekif)
		freeinfkif(ekif);
	    strcpy(errtemp, *s);
#ifdef AKII_DEBUG
	    dbgleave(dbg_readkifinfix);
#endif
	    return (NULL);
	}
	if (!kif)
	    ekif = kif = akif;
	else {
	    kif->next = akif;
	    akif->prev = kif;
	    kif = kif->next;
	}
    }
    if ((pee == PEE_PRE) || (pee == PEE_VPRE)) {
	errno = EVEGEREVAR;
	strcpy(errtemp, eleje);
	freeinfkif(ekif);
#ifdef AKII_DEBUG
	dbgleave(dbg_readkifinfix);
#endif
	return (NULL);
    }
    ekif = mkpostkif(ekif);
#ifdef AKII_DEBUG
    dbgleave(dbg_readkifinfix);
#endif
    return (ekif);
}
void freeinfkif(struct kif *kif)
{
    struct kif *k = kif;

#ifdef AKII_DEBUG
    dbgenter(dbg_freeinfkif, "?");
#endif
    while (kif) {
	if (kif->tipus != KIF_OP)
	    free(kif->ertek);
	if (kif->tipus == KIF_FV)
	    free(kif->param);
	k = kif->next;
	free(kif);
	kif = k;
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_freeinfkif);
#endif
    return;
}
struct kif *mkpostkif(struct kif *kif)
{
    struct kif *sor = NULL, *esor = NULL;
    struct kif *ops = NULL;

#ifdef AKII_DEBUG
    dbgenter(dbg_mkpostfix, "?");
#endif
    while (kif) {
	if (kif->tipus != KIF_OP) {
	    if (!esor) {
		sor = kif;
		kif = kif->next;
		sor->prev = sor->next = NULL;
		esor = sor;
	    }
	    else {
		sor->next = kif;
		kif = kif->next;
		sor->next->prev = sor;
		sor = sor->next;
		sor->next = NULL;
	    }
	}
	else {
	    if (!ops) {
		ops = kif;
		kif = kif->next;
		ops->prev = ops->next = NULL;
	    }
	    else {
		while ((OP(ops->ertek)->pl > OP(kif->ertek)->pl) || ((OP(ops->ertek)->pl == PL_MULTI)) && (OP(kif->ertek)->pl == PL_MULTI)) {
		    sor->next = ops;
		    ops = ops->prev;
		    sor->next->prev = sor;
		    sor = sor->next;
		    sor->next = NULL;
		    if (ops)
			ops->next = NULL;
		    else
			break;
		}
		if (!ops) {
		    ops = kif;
		    kif = kif->next;
		    ops->prev = ops->next = NULL;
		}
		else {
		    ops->next = kif;
		    kif = kif->next;
		    ops->next->prev = ops;
		    ops = ops->next;
		    ops->next = NULL;
		}
	    }
	}
    }
    while (ops) {
	if (!sor) {
	    sor = ops;
	    ops = ops->prev;
	    sor->prev = sor->next = NULL;
	    esor = sor;
	}
	else {
	    sor->next = ops;
	    ops = ops->prev;
	    sor->next->prev = sor;
	    sor = sor->next;
	    sor->next = NULL;
	}
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_mkpostfix);
#endif
    return (esor);
}
void freekiffa(struct kif *kif)
{
#ifdef AKII_DEBUG
    dbgenter(dbg_freekiffa, "?");
#endif
    if (!kif) {
#ifdef AKII_DEBUG
	dbgleave(dbg_freekiffa);
#endif
	return;
    }
    if (kif->bal)
	freekiffa(kif->bal);
    if (kif->jobb)
	freekiffa(kif->jobb);
    switch (kif->tipus) {
	case KIF_VAR:
	case KIF_CONST:
	    free(kif->ertek);
	    break;
	case KIF_FV:
	    free(kif->ertek);
	    free(kif->param);
	    break;
	case KIF_KIF:
	    freekiffa((struct kif *) kif->ertek);
	    break;
	case KIF_OP:
	    break;
    }
    free(kif);
#ifdef AKII_DEBUG
    dbgleave(dbg_freekiffa);
#endif
    return;
}
struct kif *mkkiffa(struct kif *kif)
{
    struct kif *fa;

#ifdef AKII_DEBUG
    dbgenter(dbg_mkkiffa, "?");
#endif
    while (kif) {
	if (kif->tipus == KIF_OP) {
	    if (OP(kif->ertek)->pl >= PL_UNARY) {
		if (kif->prev) {
		    kif->bal = kif->prev;
		    kif->prev = kif->prev->prev;
		    if (kif->prev)
			kif->prev->next = kif;
		}
		else {
		    errno = ENOOPER;
		    strcpy(errtemp, EEZUCCSE);
#ifdef AKII_DEBUG
		    dbgleave(dbg_mkkiffa);
#endif
		    return (NULL);
		}
	    }
	    else {
		if ((kif->prev) && (kif->prev->prev)) {
		    kif->jobb = kif->prev;
		    kif->bal = kif->prev->prev;
		    kif->prev = kif->prev->prev->prev;
		    if (kif->prev)
			kif->prev->next = kif;
		}
		else {
		    errno = ENOOPER;
		    strcpy(errtemp, EEZUCCSE);
#ifdef AKII_DEBUG
		    dbgleave(dbg_mkkiffa);
#endif
		    return (NULL);
		}
	    }
	}
	if (kif->next)
	    kif = kif->next;
	else
	    break;
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_mkkiffa);
#endif
    return (kif);
}
int isopchar(char c)
{
    switch (c) {
	case '+':
	case '-':
	case '*':
	case '/':
	case '%':
	case '~':
	case '|':
	case '&':
	case '=':
	case '!':
	case '<':
	case '>':
	    return (1);
    }
    return (0);
}

char *getoperand(struct kif *kif, int *mustfree)
{
    char *s, *q;
    struct valtozo *var;
    int save_varflag = varflag;

#ifdef AKII_DEBUG
    dbgenter(dbg_getoperand, "?");
#endif
    if (!kif) {
#ifdef AKII_DEBUG
	dbgleave(dbg_getoperand);
#endif
	return (NULL);
    }
    switch (kif->tipus) {
	case KIF_CONST:
	    *mustfree = 1;
	    q = kif->ertek;
	    dontquote = 1;
	    s = parsedollar(q);
	    dontquote = 0;
	    if (errno) {
		free(s);
		s = NULL;
	    }
#ifdef AKII_DEBUG
	    dbgleave(dbg_getoperand);
#endif
	    return (s);
	case KIF_VAR:
	    s = kif->ertek;
	    varflag = VFLAG_NORMAL;
	    var = getvar(&varlist, &s, 0, 1);
	    varflag = save_varflag;
	    if (!var) {
		*mustfree = 1;
		s = malloc(1);
		*s = 0;
#ifdef AKII_DEBUG
		dbgleave(dbg_getoperand);
#endif
		return (s);
	    }
	    if (var->tipus != VAR_NORMAL) {
		if (var->tipus == VAR_STRUCT)
		    errno = ESTRUCT;
		else if (var->tipus != VAR_VEZER)
		    errno = EEZFV;
		else
		    errno = EEZVEZER;
		strcpy(errtemp, kif->ertek);
#ifdef AKII_DEBUG
		dbgleave(dbg_getoperand);
#endif
		return (NULL);
	    }
	    *mustfree = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg_getoperand);
#endif
	    return (var->ertek);
	case KIF_KIF:
	    s = dokif((struct kif *) kif->ertek);
	    *mustfree = 1;
#ifdef AKII_DEBUG
	    dbgleave(dbg_getoperand);
#endif
	    return (s);
	case KIF_OP:
#ifdef AKII_DEBUG
	    dbgenter(dbg__valamilyen_op, "?");
#endif
	    s = (OP(kif->ertek)->fv) (kif);
	    if (errno == EINVAL)
		errno = 0;
#ifdef AKII_DEBUG
	    dbgleave(dbg__valamilyen_op);
#endif
	    *mustfree = 1;
#ifdef AKII_DEBUG
	    dbgleave(dbg_getoperand);
#endif
	    return (s);
	case KIF_FV:
	    s = kif->ertek;
	    var = getvar(&varlist, &s, 0, 0);
	    *mustfree = 1;
	    if (!var) {
		errno = EEZNEMFV;
		strcpy(errtemp, kif->ertek);
#ifdef AKII_DEBUG
		dbgleave(dbg_getoperand);
#endif
		return (NULL);
	    }
	    if ((var->tipus != VAR_FV) && (var->tipus != VAR_ALIAS)) {
		errno = EEZNEMFV;
		strcpy(errtemp, kif->ertek);
#ifdef AKII_DEBUG
		dbgleave(dbg_getoperand);
#endif
		return (NULL);
	    }
	    s = parsedollar(kif->param);
	    if (errno) {
		free(s);
#ifdef AKII_DEBUG
		dbgleave(dbg_getoperand);
#endif
		return (NULL);
	    }
	    q = callfv(var, s);
	    free(s);
	    if (errno) {
#ifdef AKII_DEBUG
		dbgleave(dbg_getoperand);
#endif
		return (NULL);
	    }
	    s = malloc(strlen(q) + 1);
	    strcpy(s, q);
#ifdef AKII_DEBUG
	    dbgleave(dbg_getoperand);
#endif
	    return (s);
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_getoperand);
#endif
    return (NULL);
}

int dtos(double f, char *s)
{
    char temp[128];
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_dtos, "?");
#endif
    if (f < -1e20) {
	strcpy(s, "<LIMITED>");
#ifdef AKII_DEBUG
	dbgleave(dbg_dtos);
#endif
	return (1);
    }
    if (DF(f) == 0.0) {
	sprintf(s, "%ld", (long) f);
#ifdef AKII_DEBUG
	dbgleave(dbg_dtos);
#endif
	return (1);
    }
    sprintf(temp, "%.10lf", DF(f));
    for (i = 11; i > 0; i--) {
	if (temp[i] != '0') {
	    if (temp[i] == '.')
		temp[i] = 0;
	    break;
	}
	temp[i] = 0;
    }
    sprintf(s, "%ld%s", (long) f, temp + ((f >= 0) ? 1 : 2));
#ifdef AKII_DEBUG
    dbgleave(dbg_dtos);
#endif
    return (0);
}

void varcopy(struct valtozo *ide, struct valtozo *ezt)
{
    struct valtozo *var, *v, *v2;
    struct aliasstr *ali, *a;
    char *s;
    char temp[300];
    int i;
    int save_varflag = varflag;

    if (ide->tipus == VAR_STRUCT && ide->ertek)
	freevarlist((struct valtozo *) ide->ertek, 1);
    else if (ide->tipus == VAR_NORMAL)
	free(ide->ertek);
    else
	return;
    ide->tipus = VAR_STRUCT;
    ide->ertek = NULL;
    var = (struct valtozo *) ezt->ertek;
    while (var) {
	v = var->next;
	strcpy(temp, var->nev);
	s = temp;
	varflag = ide->flag;
	v2 = getvar((struct valtozo **) &ide->ertek, &s, 1, 1);
	varflag = save_varflag;
	switch (var->tipus) {
	    case VAR_NORMAL:
		setvar(v2, var->ertek);
		break;
	    case VAR_STRUCT:
		varcopy(v2, var);
		break;
	    case VAR_ALIAS:
		v2->tipus = VAR_ALIAS;
		free(v2->ertek);
		ali = (struct aliasstr *) var->ertek;
		a = malloc(sizeof(struct aliasstr));

		a->cpnum = ali->cpnum;
		a->cps = malloc(a->cpnum * sizeof(char *));

		for (i = 0; i < a->cpnum; i++)
		    a->cps[i] = strdup(ali->cps[i]);
		a->par = ali->par;
		a->par->refnum++;
		v2->ertek = (char *) a;
		break;
	    case VAR_FV:
	    case VAR_VEZER:
	    case VAR_FOREACH:
		break;
	}
	var = v;
    }
    return;
}
char *op_eq(struct kif *kif)
{
    char *ret;
    char *sj;
    struct valtozo *var, *structcopy = NULL;
    int jmustfree;
    int save_varflag = varflag;

    if (!kif)
	return (NULL);
    sj = getoperand(kif->jobb, &jmustfree);
    if (!sj) {
	if (errno != ESTRUCT)
	    return (NULL);
	ret = kif->jobb->ertek;
	varflag = VFLAG_NORMAL;
	var = getvar(&varlist, &ret, 1, 1);
	varflag = save_varflag;
	if (var->tipus != VAR_STRUCT)
	    return (NULL);
	structcopy = var;
    }
    ret = kif->bal->ertek;
    varflag = VFLAG_NORMAL;
    var = getvar(&varlist, &ret, 1, 0);
    varflag = save_varflag;
    if (!var)
	return (NULL);
    if (structcopy && (var->tipus == VAR_NORMAL || var->tipus == VAR_STRUCT)) {
	varcopy(var, structcopy);
	ret = malloc(1);
	*ret = 0;
	return (ret);
    }
    if (var->tipus != VAR_NORMAL) {
	if (jmustfree)
	    free(sj);
	if (var->tipus != VAR_VEZER)
	    errno = EEZFV;
	else
	    errno = EEZVEZER;
	strcpy(errtemp, var->nev);
	return (NULL);
    }
    setvar(var, sj);
    if (!jmustfree) {
	ret = malloc(strlen(sj) + 1);
	strcpy(ret, sj);
    }
    else
	ret = sj;
    return (ret);
}
char *op_valami_eq(struct kif *kif, char valami)
{
    char ret[124];
    char *t;
    char *sj;
    struct valtozo *var;
    int jmustfree;
    double lj, lb;
    int save_varflag = varflag;

    if (!kif)
	return (NULL);
    sj = getoperand(kif->jobb, &jmustfree);
    if (!sj)
	return (NULL);
    lj = strtod(sj, &t);
    if (*t)
	lj = 0;
    if (jmustfree)
	free(sj);
    sj = kif->bal->ertek;
    varflag = VFLAG_NORMAL;
    var = getvar(&varlist, &sj, 1, 0);
    varflag = save_varflag;
    if (!var)
	return (NULL);
    if (var->tipus != VAR_NORMAL) {
	if (jmustfree)
	    free(sj);
	if (var->tipus != VAR_VEZER)
	    errno = EEZFV;
	else
	    errno = EEZVEZER;
	strcpy(errtemp, var->nev);
	return (NULL);
    }
    lb = strtod(var->ertek, &t);
    if (*t)
	lb = 0;
    switch (valami) {
	case '+':
	    lb += lj;
	    break;
	case '-':
	    lb -= lj;
	    break;
	case '*':
	    lb *= lj;
	    break;
	case '/':
	    if (lj != 0)
		lb /= lj;
	    else
		errno = EOSZTNULL;
	    break;
	case '%':
	    if (lj != 0)
		lb = (long) lb % (long) lj;
	    else
		errno = EOSZTNULL;
	    break;
    }
    dtos(lb, ret);
    setvar(var, ret);
    sj = malloc(strlen(ret) + 1);
    strcpy(sj, ret);
    if (errno) {
	free(sj);
	return (NULL);
    }
    return (sj);
}
char *op_plus_eq(struct kif *kif)
{
    return (op_valami_eq(kif, '+'));
}
char *op_minus_eq(struct kif *kif)
{
    return (op_valami_eq(kif, '-'));
}
char *op_times_eq(struct kif *kif)
{
    return (op_valami_eq(kif, '*'));
}
char *op_slash_eq(struct kif *kif)
{
    return (op_valami_eq(kif, '/'));
}
char *op_mod_eq(struct kif *kif)
{
    return (op_valami_eq(kif, '%'));
}
char *op_cat_eq(struct kif *kif)
{
    char *ret;
    char *sj;
    struct valtozo *var;
    int jmustfree;
    int save_varflag = varflag;

    if (!kif)
	return (NULL);
    sj = getoperand(kif->jobb, &jmustfree);
    if (!sj)
	return (NULL);
    ret = kif->bal->ertek;
    varflag = VFLAG_NORMAL;
    var = getvar(&varlist, &ret, 1, 0);
    varflag = save_varflag;
    if (!var)
	return (NULL);
    if (var->tipus != VAR_NORMAL) {
	if (jmustfree)
	    free(sj);
	if (var->tipus != VAR_VEZER)
	    errno = EEZFV;
	else
	    errno = EEZVEZER;
	strcpy(errtemp, var->nev);
	return (NULL);
    }
    ret = malloc(strlen(var->ertek) + strlen(sj) + 1);
    strcpy(ret, var->ertek);
    strcat(ret, sj);
    setvar(var, ret);
    free(ret);
    if (jmustfree)
	free(sj);
    ret = malloc(strlen(var->ertek) + 1);
    strcpy(ret, var->ertek);
    return (ret);
}
char *op_orand(struct kif *kif, char valami)
{
    char *ret;
    char *t;
    char *sj, *sb;
    double lj, lb;
    int jmustfree, bmustfree;

    if (!kif)
	return (NULL);
    sb = getoperand(kif->bal, &bmustfree);
    if (!sb)
	return (NULL);
    lb = strtod(sb, &t);
    if (*t)
	lb = 0;
    if (bmustfree)
	free(sb);
    ret = malloc(2);
    ret[1] = 0;
    if (lb != 0)
	ret[0] = '1';
    else
	ret[0] = '0';
    if (((valami == '|') && (ret[0] == '1')) || ((valami == '&') && (ret[0] == '0')))
	return (ret);
    sj = getoperand(kif->jobb, &jmustfree);
    if (!sj)
	return (NULL);
    lj = strtod(sj, &t);
    if (*t)
	lj = 0;
    if (jmustfree)
	free(sj);
    if (lj != 0)
	ret[0] = '1';
    else
	ret[0] = '0';
    return (ret);
}
char *op_or(struct kif *kif)
{
    return (op_orand(kif, '|'));
}
char *op_and(struct kif *kif)
{
    return (op_orand(kif, '&'));
}
char *op_eqnoteq(struct kif *kif, char ifeq, char ifnoteq)
{
    char *ret;
    char *sj, *sb, *t;
    int jmustfree, bmustfree;
    double lb, lj;
    int szam = 1;

    if (!kif)
	return (NULL);
    sj = getoperand(kif->jobb, &jmustfree);
    if (!sj)
	return (NULL);
    sb = getoperand(kif->bal, &bmustfree);
    if (!sb)
	return (NULL);
    ret = malloc(2);
    ret[1] = 0;
    lb = strtod(sb, &t);
    if (*t)
	szam = 0;
    lj = strtod(sj, &t);
    if (*t)
	szam = 0;
    if (!szam) {
	if (istrcmp(sb, sj) == 0)
	    ret[0] = ifeq;
	else
	    ret[0] = ifnoteq;
    }
    else {
	if (lb == lj)
	    ret[0] = ifeq;
	else
	    ret[0] = ifnoteq;
    }
    if (jmustfree)
	free(sj);
    if (bmustfree)
	free(sb);
    return (ret);
}
char *op_is_eq(struct kif *kif)
{
    return (op_eqnoteq(kif, '1', '0'));
}
char *op_is_noteq(struct kif *kif)
{
    return (op_eqnoteq(kif, '0', '1'));
}
char *op_rel(struct kif *kif, int miez)
{
    char *ret;
    char *t;
    char *sj, *sb;
    double lj, lb;
    int jmustfree, bmustfree;

    if (!kif)
	return (NULL);
    sj = getoperand(kif->jobb, &jmustfree);
    if (!sj)
	return (NULL);
    sb = getoperand(kif->bal, &bmustfree);
    if (!sb)
	return (NULL);
    lj = strtod(sj, &t);
    if (*t)
	lj = 0;
    lb = strtod(sb, &t);
    if (*t)
	lb = 0;
    ret = malloc(2);
    ret[1] = 0;
    switch (miez) {
	case 1:
	    if (lb < lj)
		ret[0] = '1';
	    else
		ret[0] = '0';
	    break;
	case 2:
	    if (lb > lj)
		ret[0] = '1';
	    else
		ret[0] = '0';
	    break;
	case 3:
	    if (lb <= lj)
		ret[0] = '1';
	    else
		ret[0] = '0';
	    break;
	case 4:
	    if (lb >= lj)
		ret[0] = '1';
	    else
		ret[0] = '0';
	    break;
    }
    if (jmustfree)
	free(sj);
    if (bmustfree)
	free(sb);
    return (ret);
}
char *op_less(struct kif *kif)
{
    return (op_rel(kif, 1));
}
char *op_greater(struct kif *kif)
{
    return (op_rel(kif, 2));
}
char *op_less_eq(struct kif *kif)
{
    return (op_rel(kif, 3));
}
char *op_greater_eq(struct kif *kif)
{
    return (op_rel(kif, 4));
}
char *op_valami(struct kif *kif, char valami)
{
    char *ret;
    char *t;
    char *sj, *sb;
    double lj, lb;
    int jmustfree, bmustfree;

    if (!kif)
	return (NULL);
    sj = getoperand(kif->jobb, &jmustfree);
    if (!sj)
	return (NULL);
    sb = getoperand(kif->bal, &bmustfree);
    if (!sb)
	return (NULL);
    lj = strtod(sj, &t);
    if (*t)
	lj = 0;
    lb = strtod(sb, &t);
    if (*t)
	lb = 0;
    ret = malloc(128);
    switch (valami) {
	case '+':
	    lb += lj;
	    break;
	case '-':
	    lb -= lj;
	    break;
	case '*':
	    lb *= lj;
	    break;
	case '/':
	    if (lj != 0)
		lb /= lj;
	    else
		errno = EOSZTNULL;
	    break;
	case ':':
	    if (lj != 0)
		lb = (int) (lb / lj);
	    else
		errno = EOSZTNULL;
	    break;
	case '%':
	    if (lj != 0)
		lb = (long) lb % (long) lj;
	    else
		errno = EOSZTNULL;
	    break;
    }
    dtos(lb, ret);
    ret = realloc(ret, strlen(ret) + 1);
    if (jmustfree)
	free(sj);
    if (bmustfree)
	free(sb);
    if (errno) {
	free(ret);
	return (NULL);
    }
    return (ret);
}
char *op_plus(struct kif *kif)
{
    return (op_valami(kif, '+'));
}
char *op_minus(struct kif *kif)
{
    return (op_valami(kif, '-'));
}
char *op_cat(struct kif *kif)
{
    char *ret;
    char *sj, *sb;
    int jmustfree, bmustfree;

    if (!kif)
	return (NULL);
    sj = getoperand(kif->jobb, &jmustfree);
    if (!sj)
	return (NULL);
    sb = getoperand(kif->bal, &bmustfree);
    if (!sb)
	return (NULL);
    ret = malloc(strlen(sj) + strlen(sb) + 1);
    strcpy(ret, sb);
    strcat(ret, sj);
    if (jmustfree)
	free(sj);
    if (bmustfree)
	free(sb);
    return (ret);
}
char *op_times(struct kif *kif)
{
    return (op_valami(kif, '*'));
}
char *op_slash(struct kif *kif)
{
    return (op_valami(kif, '/'));
}
char *op_slash_egesz(struct kif *kif)
{
    return (op_valami(kif, ':'));
}
char *op_mod(struct kif *kif)
{
    return (op_valami(kif, '%'));
}
char *op_u_valami(struct kif *kif, int valami)
{
    char *ret;
    char *t;
    char *sb;
    double lb;
    int bmustfree;

    if (!kif)
	return (NULL);
    sb = getoperand(kif->bal, &bmustfree);
    if (!sb)
	return (NULL);
    lb = strtod(sb, &t);
    if (*t)
	lb = 0;
    ret = malloc(256);
    if (valami == 2)
	lb = -lb;
    dtos(lb, ret);
    ret = realloc(ret, strlen(ret) + 1);
    if (bmustfree)
	free(sb);
    return (ret);
}
char *op_u_plus(struct kif *kif)
{
    return (op_u_valami(kif, 1));
}
char *op_u_minus(struct kif *kif)
{
    return (op_u_valami(kif, 2));
}
char *op_u_not(struct kif *kif)
{
    char *ret;
    char *t;
    char *sb;
    double lb;
    int bmustfree;

    if (!kif)
	return (NULL);
    sb = getoperand(kif->bal, &bmustfree);
    if (!sb)
	return (NULL);
    lb = strtod(sb, &t);
    if (*t)
	lb = 0;
    ret = malloc(2);
    ret[1] = 0;
    if (lb != 0)
	ret[0] = '0';
    else
	ret[0] = '1';
    if (bmustfree)
	free(sb);
    return (ret);
}
char *op_u_xx(struct kif *kif, int xx)
{
    char ret[256];
    char *t;
    struct valtozo *var;
    char *sj;
    double lb;
    int save_varflag = varflag;

    if (!kif)
	return (NULL);
    sj = kif->bal->ertek;
    varflag = VFLAG_NORMAL;
    var = getvar(&varlist, &sj, 1, 0);
    varflag = save_varflag;
    if (!var)
	return (NULL);
    if (var->tipus != VAR_NORMAL) {
	if (var->tipus != VAR_VEZER)
	    errno = EEZFV;
	else
	    errno = EEZVEZER;
	strcpy(errtemp, var->nev);
	return (NULL);
    }
    lb = strtod(var->ertek, &t);
    if (*t)
	lb = 0;
    if (((int) kif->param) == BUNARY_POST) {
	dtos(lb, ret);
	sj = malloc(strlen(ret) + 1);
	strcpy(sj, ret);
    }
    if (xx == 1)
	lb++;
    else
	lb--;
    dtos(lb, ret);
    setvar(var, ret);
    if (((int) kif->param) == BUNARY_PRE) {
	sj = malloc(strlen(ret) + 1);
	strcpy(sj, ret);
    }
    return (sj);
}
char *op_u_plusplus(struct kif *kif)
{
    return (op_u_xx(kif, 1));
}
char *op_u_minusminus(struct kif *kif)
{
    return (op_u_xx(kif, 2));
}
